-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core for SILC V2
--! @project    CAESAR Candidate Evaluation
--! @author     
--! @copyright  
--!             
--!             
--! @license    
--!             
--!             
--!             
--! @note       
--!             
--!             
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use work.AEAD_pkg.ALL;
--use work.AES_pkg.ALL;

use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;



entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
--        G_DBLK_SIZE     : integer := 128;    --! Data   Block size n(bits) : 128(AES128)/64(TWINE80)
--        G_KEY_SIZE      : integer := 128;    --! Key    Key length(bits)   : 128(AES128)/80(TWINE80)
--        G_TAG_SIZE      : integer := 64;     --! Tag
        G_DBLK_SIZE     : integer := 64;    --! Data   Block size n(bits) : 128(AES128)/64(TWINE80)
        G_KEY_SIZE      : integer := 80;    --! Key    Key length(bits)   : 128(AES128)/80(TWINE80)
        G_TAG_SIZE      : integer := 32;     --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
--        G_LBS_BYTES     : integer := 4;
        G_LBS_BYTES     : integer := 3;
        G_MAX_LEN       : integer := SINGLE_PASS_MAX;        --(+)caesar_hw_devpkg-1.0-3

--v2(+) from
        G_PARAM         : std_logic_vector(7 downto 0)    := X"CC"    --param (AES-128, Nonce=12bytes, Tag= 8bytes)
--        G_PARAM         : std_logic_vector(7 downto 0)    := X"C0"    --param (AES-128, Nonce=12bytes, Tag= 8bytes)
--        G_PARAM       : std_logic_vector(7 downto 0)    := X"D0"    --param (AES-128, Nonce= 8bytes, Tag= 8bytes)
--v2(+) to
    );

    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;        -- '0' : ENC / '1' : DEC
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );

end entity CipherCore;

architecture structure of CipherCore is

	component cloc_twine80_aeadif
	port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;        -- '0' : ENC / '1' : DEC
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
	);
	end component;

begin

	u0: cloc_twine80_aeadif port map (
        --! Global
        clk             => clk,
        rst             => rst,
        --! PreProcessor (data)
        key             => key,
        bdi             => bdi,
        --! PreProcessor (controls)
        key_ready       => key_ready,
        key_valid       => key_valid,
        key_update      => key_update,
        decrypt         => decrypt,
        bdi_ready       => bdi_ready,
        bdi_valid       => bdi_valid,
        bdi_type        => bdi_type,
        bdi_partial     => bdi_partial,
        bdi_eot         => bdi_eot,
        bdi_eoi         => bdi_eoi,
        bdi_size        => bdi_size,
        bdi_valid_bytes => bdi_valid_bytes,
        bdi_pad_loc     => bdi_pad_loc,
        --! PostProcessor
        bdo             => bdo,
        bdo_valid       => bdo_valid,
        bdo_ready       => bdo_ready,
        bdo_size        => bdo_size,
        msg_auth_done   => msg_auth_done,
        msg_auth_valid  => msg_auth_valid
	);

end structure;
