------------------------------------------------------------------------------
-- CLOC-Twine80 with AEAD I/F
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity cloc_twine80_aeadif is
	port (
		bdi				: in  std_logic_vector(63 downto 0);
		bdi_valid		: in  std_logic;
		bdi_ready		: out std_logic;
		decrypt			: in  std_logic;
		bdi_type		: in  std_logic_vector(2 downto 0);
		bdi_partial		: in  std_logic;
		bdi_eot			: in  std_logic;
		bdi_eoi			: in  std_logic;
		bdi_size		: in  std_logic_vector(3 downto 0);
		bdi_valid_bytes	: in  std_logic_vector(7 downto 0);
		bdi_pad_loc		: in  std_logic_vector(7 downto 0);

		key				: in  std_logic_vector(79 downto 0);
		key_valid		: in  std_logic;
		key_update		: in  std_logic;
		key_ready		: out std_logic;

		bdo				: out std_logic_vector(63 downto 0);
		bdo_valid		: out std_logic;
		bdo_ready		: in  std_logic;
		bdo_size		: out std_logic_vector(3 downto 0);
		msg_auth_done	: out std_logic;
		msg_auth_valid	: out std_logic;

		rst				: in  std_logic;	-- sync reset, active H
		clk				: in  std_logic
	);
end cloc_twine80_aeadif;

architecture RTL of cloc_twine80_aeadif is

	component twine80
	port (
		din		: in  std_logic_vector(63 downto 0);
		dout	: out std_logic_vector(63 downto 0);
		kin		: in  std_logic_vector(79 downto 0);
		start	: in  std_logic;
		busy	: out std_logic;
		rst		: in  std_logic;
		clk		: in  std_logic
	);
	end component;

	constant PARAM	: std_logic_vector(7 downto 0)  := To_StdLogicVector(X"CC");		-- twine t6n4
	constant TAU	: std_logic_vector(31 downto 0) := CONV_STD_LOGIC_VECTOR(4,32);

	------------------------------------------------------------/
	-- functions
	------------------------------------------------------------/

	-- ozp
	function ozp_func (
		a		: std_logic_vector(63 downto 0);
		len		: std_logic_vector(3 downto 0);	-- valid length (unit:byte, 0-8)
		thru	: std_logic						-- force thru
	) return std_logic_vector is
		variable len_r				: std_logic_vector(3 downto 0);
		variable and_mask, or_mask	: std_logic_vector(63 downto 0);
		variable ozp_func_ret		: std_logic_vector(63 downto 0);
	begin
		if (thru = '1') then
			len_r	:= CONV_STD_LOGIC_VECTOR(8,4);
		else
			len_r	:= len;
		end if;

		case (len_r) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			and_mask	:= To_StdLogicVector(X"0000000000000000");
			or_mask		:= To_StdLogicVector(X"8000000000000000");
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			and_mask	:= To_StdLogicVector(X"FF00000000000000");
			or_mask		:= To_StdLogicVector(X"0080000000000000");
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			and_mask	:= To_StdLogicVector(X"FFFF000000000000");
			or_mask		:= To_StdLogicVector(X"0000800000000000");
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFF0000000000");
			or_mask		:= To_StdLogicVector(X"0000008000000000");
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFF00000000");
			or_mask		:= To_StdLogicVector(X"0000000080000000");
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFF000000");
			or_mask		:= To_StdLogicVector(X"0000000000800000");
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFF0000");
			or_mask		:= To_StdLogicVector(X"0000000000008000");
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFF00");
			or_mask		:= To_StdLogicVector(X"0000000000000080");
		when others =>	-- >=8; thru
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFFFF");
			or_mask		:= To_StdLogicVector(X"0000000000000000");
		end case;

		ozp_func_ret	:= (a and and_mask) or or_mask;
		return ozp_func_ret;
	end ozp_func;

	-- fix0
	function fix0_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable fix0_func_ret	: std_logic_vector(63 downto 0);
	begin
		fix0_func_ret	:= ('0' & a(62 downto 0));
		return fix0_func_ret;
	end fix0_func;

	-- fix1
	function fix1_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable fix1_func_ret	: std_logic_vector(63 downto 0);
	begin
		fix1_func_ret	:= ('1' & a(62 downto 0));
		return fix1_func_ret;
	end fix1_func;

	-- f1
	function f1_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable w4, w3, w2, w1, t0, t1	: std_logic_vector(15 downto 0);
		variable f1_func_ret			: std_logic_vector(63 downto 0);
	begin
		w1	:= a(63 downto 48);
		w2	:= a(47 downto 32);
		w3	:= a(31 downto 16);
		w4	:= a(15 downto 0);
		t0	:= w1 xor w3;
		t1	:= w2 xor w4;
		f1_func_ret	:= (t0 & t1 & (t0 xor w2) & (t1 xor w3));
		return f1_func_ret;
	end f1_func;

	-- f2
	function f2_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable w4, w3, w2, w1	: std_logic_vector(15 downto 0);
		variable f2_func_ret	: std_logic_vector(63 downto 0);
	begin
		w1	:= a(63 downto 48);
		w2	:= a(47 downto 32);
		w3	:= a(31 downto 16);
		w4	:= a(15 downto 0);
		f2_func_ret	:= (w2 & w3 & w4 & (w1 xor w2));
		return f2_func_ret;
	end f2_func;

	-- g1
	function g1_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable w4, w3, w2, w1	: std_logic_vector(15 downto 0);
		variable g1_func_ret	: std_logic_vector(63 downto 0);
	begin
		w1	:= a(63 downto 48);
		w2	:= a(47 downto 32);
		w3	:= a(31 downto 16);
		w4	:= a(15 downto 0);
		g1_func_ret	:= (w3 & w4 & (w1 xor w2) & (w2 xor w3));
		return g1_func_ret;
	end g1_func;

	-- g2
	function g2_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable w4, w3, w2, w1	: std_logic_vector(15 downto 0);
		variable g2_func_ret	: std_logic_vector(63 downto 0);
	begin
		w1	:= a(63 downto 48);
		w2	:= a(47 downto 32);
		w3	:= a(31 downto 16);
		w4	:= a(15 downto 0);
		g2_func_ret	:= (w2 & w3 & w4 & (w1 xor w2));
		return g2_func_ret;
	end g2_func;

	-- h
	function h_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable w4, w3, w2, w1, t0	: std_logic_vector(15 downto 0);
		variable h_func_ret			: std_logic_vector(63 downto 0);
	begin
		w1	:= a(63 downto 48);
		w2	:= a(47 downto 32);
		w3	:= a(31 downto 16);
		w4	:= a(15 downto 0);
		t0	:= w1 xor w2;
		h_func_ret	:= (t0 & (w2 xor w3) & (w3 xor w4) & (t0 xor w4));
		return h_func_ret;
	end h_func;

	-- msb
	function msb_func (
		a		: std_logic_vector(63 downto 0);
		len		: std_logic_vector(3 downto 0);	-- valid length (unit:byte, 0-8)
		thru	: std_logic						-- force thru
	) return std_logic_vector is
		variable len_r			: std_logic_vector(3 downto 0);
		variable and_mask		: std_logic_vector(63 downto 0);
		variable msb_func_ret	: std_logic_vector(63 downto 0);
	begin
		if (thru = '1') then
			len_r	:= CONV_STD_LOGIC_VECTOR(8,4);
		else
			len_r	:= len;
		end if;

		case (len_r) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			and_mask	:= To_StdLogicVector(X"0000000000000000");
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			and_mask	:= To_StdLogicVector(X"FF00000000000000");
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			and_mask	:= To_StdLogicVector(X"FFFF000000000000");
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFF0000000000");
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFF00000000");
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFF000000");
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFF0000");
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFF00");
		when others =>	-- >=8; thru
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFFFF");
		end case;

		msb_func_ret	:= a and and_mask;
		return msb_func_ret;
	end msb_func;

	-- mux
	function mux10_func (
		a0	: std_logic_vector(63 downto 0);
		a1	: std_logic_vector(63 downto 0);
		a2	: std_logic_vector(63 downto 0);
		a3	: std_logic_vector(63 downto 0);
		a4	: std_logic_vector(63 downto 0);
		a5	: std_logic_vector(63 downto 0);
		a6	: std_logic_vector(63 downto 0);
		a7	: std_logic_vector(63 downto 0);
		a8	: std_logic_vector(63 downto 0);
		a9	: std_logic_vector(63 downto 0);
		a10	: std_logic_vector(63 downto 0);
		a11	: std_logic_vector(63 downto 0);
		sel	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable mux10_func_ret	: std_logic_vector(63 downto 0);
	begin
		case (sel) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			mux10_func_ret	:= a0;
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			mux10_func_ret	:= a1;
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			mux10_func_ret	:= a2;
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			mux10_func_ret	:= a3;
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			mux10_func_ret	:= a4;
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			mux10_func_ret	:= a5;
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			mux10_func_ret	:= a6;
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			mux10_func_ret	:= a7;
		when CONV_STD_LOGIC_VECTOR(8,4) =>
			mux10_func_ret	:= a8;
		when CONV_STD_LOGIC_VECTOR(9,4) =>
			mux10_func_ret	:= a9;
		when CONV_STD_LOGIC_VECTOR(10,4) =>
			mux10_func_ret	:= a10;
		when others =>
			mux10_func_ret	:= a11;
		end case;
		return mux10_func_ret;
	end mux10_func;

	------------------------------------------------------------/
	-- data path
	------------------------------------------------------------/
	signal nonce_reg					: std_logic_vector(55 downto 0);	-- NOTE; TAU==4
	signal key_reg						: std_logic_vector(79 downto 0);
	signal v_reg, x_reg, y_reg, o_reg, od_reg	: std_logic_vector(63 downto 0);
	signal alen_reg, mlen_reg, nlen_reg	: std_logic_vector(3 downto 0);
	signal msb_a_reg, msb_ozp_a_reg		: std_logic;

	signal ek_din, ek_dout, ek_dout_h	: std_logic_vector(63 downto 0);
	signal ek_start_reg					: std_logic;
	signal ek_busy						: std_logic;

	-- Ei
	signal ei_sel_reg					: std_logic_vector(3 downto 0);	-- data path control
	signal es0_reg, es1_reg, es2_reg	: std_logic;	-- data path control
	signal ei0, ei1, ei2, ei3, ei4, ei5, ei6, ei7, ei8, ei9, ei10, ei11	: std_logic_vector(63 downto 0);
	signal nd0, nd1, nd2				: std_logic_vector(63 downto 0);

	-- Vi
	signal vi_sel_reg, vs0_reg			: std_logic;	-- data path control
	signal pi_par						: std_logic_vector(63 downto 0);
	signal vi0, vi1						: std_logic_vector(63 downto 0);
	signal nd10, nd11					: std_logic_vector(63 downto 0);
	signal vi_node						: std_logic_vector(63 downto 0);

	-- Oi
	signal os0_reg						: std_logic;	-- data path control
	signal oi_node						: std_logic_vector(63 downto 0);

	-- T
	signal t_node						: std_logic_vector(31 downto 0);	-- TAU == 4

	------------------------------------------------------------/
	-- control
	------------------------------------------------------------/
	signal bdi_ready_reg, key_ready_reg, bdo_valid_reg	: std_logic;
	signal enc_dec_reg					: std_logic;
	signal a1_flag_reg					: std_logic;

	signal bdo_size_reg					: std_logic_vector(3 downto 0);
	signal msg_auth_done_reg			: std_logic;
	signal msg_auth_valid_reg			: std_logic;
	signal eot_reg, eoi_reg				: std_logic;

	constant ST_WAIT	: std_logic_vector(7 downto 0) := To_StdLogicVector(X"00");
	constant ST_K0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"10");
	constant ST_N0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"20");
	constant ST_H0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"30");
	constant ST_H1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"31");
	constant ST_H2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"32");
	constant ST_H3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"33");
	constant ST_H4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"34");
	constant ST_H5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"35");
	constant ST_H6		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"36");
	constant ST_H7		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"37");
	constant ST_H8		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"38");
	constant ST_H9		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"39");
	constant ST_H10		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"40");
	constant ST_H11		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"41");
	constant ST_C0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"50");
	constant ST_C1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"51");
	constant ST_C2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"52");
	constant ST_C3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"53");
	constant ST_C4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"54");
	constant ST_C5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"55");
	constant ST_C6		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"56");
	constant ST_C7		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"57");
	constant ST_C8		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"58");
	constant ST_C9		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"59");
	constant ST_C10		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"60");
	constant ST_C11		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"61");
	constant ST_C12		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"62");
	constant ST_P0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"70");
	constant ST_P1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"71");
	constant ST_P2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"72");
	constant ST_P3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"73");
	constant ST_P4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"74");
	constant ST_P5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"75");

	signal state_reg	: std_logic_vector(7 downto 0);

begin
	------------------------------------------------------------/
	-- data path
	------------------------------------------------------------/
	ei0		<= v_reg;
	ei1		<= g1_func(v_reg);
	ei2		<= g2_func(v_reg);
	ei3		<= fix1_func(o_reg);

	nd0		<= ozp_func(bdi, alen_reg(3 downto 0), es0_reg);			-- es0_reg == 1 ... thru
	ei4		<= fix0_func(nd0);
	ei5		<= fix1_func(nd0);
	nd1		<= ek_dout	when (es1_reg = '1') else ek_dout_h;
	ei6		<= nd0 xor nd1;

	nd2		<= ozp_func(o_reg, mlen_reg(3 downto 0), es2_reg);			-- es2_reg == 1 ... thru
	ei7		<= nd2 xor x_reg;
	ei8		<= f1_func(ei7);
	ei9		<= f2_func(ei7);

	ei10	<= y_reg;
	ei11	<= y_reg xor nd0;

	ek_din	<= mux10_func(ei0, ei1, ei2, ei3, ei4, ei5, ei6, ei7, ei8, ei9, ei10, ei11, ei_sel_reg);

	pi_par	<= (PARAM & nonce_reg);										-- NOTE; TAU==4
	nd10	<= ek_dout	when (vs0_reg = '1') else ek_dout_h;
	nd11	<= nd10 xor ozp_func(pi_par, nlen_reg(3 downto 0), '0');
	vi0		<= f1_func(nd11);
	vi1		<= f2_func(nd11);
	vi_node	<= vi0	when (vi_sel_reg = '0') else vi1;

	oi_node	<= bdi xor msb_func(y_reg, mlen_reg(3 downto 0), os0_reg);	-- os0_reg == 1 ... thru

	t_node	<= x_reg(63 downto 32);

	-- Twine
	u0: twine80
	port map (
		din		=> ek_din,
		dout	=> ek_dout,
		kin		=> key_reg,
		start	=> ek_start_reg,
		busy	=> ek_busy,
		rst		=> rst,
		clk		=> clk
	);

	ek_dout_h	<= h_func(ek_dout);

	------------------------------------------------------------/
	-- control
	------------------------------------------------------------/
	p0:process (clk)
	begin
		if (clk'event and clk = '1') then
			if (rst = '1') then
				bdi_ready_reg		<= '0';
				key_ready_reg		<= '0';
				bdo_valid_reg		<= '0';
				ek_start_reg		<= '0';

				bdo_size_reg		<= CONV_STD_LOGIC_VECTOR(0,4);
				msg_auth_done_reg	<= '0';
				msg_auth_valid_reg	<= '0';
				eot_reg				<= '0';
				eoi_reg				<= '0';

				ei_sel_reg			<= CONV_STD_LOGIC_VECTOR(0,4);
				es0_reg				<= '0';
				es1_reg				<= '0';
				es2_reg				<= '0';
				vi_sel_reg			<= '0';
				vs0_reg				<= '0';
				os0_reg				<= '0';

				state_reg			<= ST_WAIT;

			else
				case (state_reg) is
				--------------------------------
				-- wait trigger
				--------------------------------
				when ST_WAIT =>
					if (key_update = '1') then
						if (key_valid = '1') then
							key_reg			<= key;
							key_ready_reg	<= '1';
							state_reg		<= ST_K0;
						end if;

					elsif (bdi_valid = '1' and bdi_type = CONV_STD_LOGIC_VECTOR(6,3)) then
						nlen_reg		<= bdi_size + '1';		-- [add 1, which corresponds to length of const value]
						nonce_reg		<= bdi(63 downto 8);	-- ignore last 8bit (for adding a constant)
						enc_dec_reg		<= decrypt;
						eot_reg			<= bdi_eot;
						eoi_reg			<= bdi_eoi;

						bdi_ready_reg	<= '1';					-- terminate bus cycle (nonce)
						state_reg		<= ST_N0;

					else
						key_ready_reg	<= '0';
						bdi_ready_reg	<= '0';
					end if;

				--------------------------------
				-- SDI: Set key
				--------------------------------
				when ST_K0 =>	-- wait end of bus cycle (key)
					key_ready_reg	<= '0';
					state_reg		<= ST_WAIT;

				--------------------------------
				-- PDI1: NONCE
				--------------------------------
				when ST_N0 =>
					bdi_ready_reg	<= '0';

					if (eoi_reg = '0') then
						a1_flag_reg	<= '1';						-- A[1]
						state_reg	<= ST_H0;

					else
						alen_reg	<= CONV_STD_LOGIC_VECTOR(0,4);	-- |A| = 0, |M| = 0; skip bus wait
						v_reg		<= CONV_STD_LOGIC_VECTOR(0,64);	-- force A==0
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(0,4);	-- Ek <= V (==0)
						state_reg	<= ST_H6;						-- special case
					end if;

				--------------------------------
				-- PD2: HASH
				--------------------------------
				when ST_H0 =>	-- init; A[1] only
					if (bdi_valid = '1') then
						if (bdi_type(2 downto 1) = "01") then		-- M/C comes here, indicating |A| = 0
							-- NOTE: do not terminate this bus cycle (M/C)
							alen_reg	<= CONV_STD_LOGIC_VECTOR(0,4);
							v_reg		<= CONV_STD_LOGIC_VECTOR(0,64);		-- force A==0
							ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(0,4);		-- Ek <= V (==0)
							state_reg	<= ST_H6;							-- special case

						else
							alen_reg	<= bdi_size;						-- (normal case)
							eot_reg		<= bdi_eot;
							eoi_reg		<= bdi_eoi;

							-- reconfig data path, for loading A
							if (bdi_eot = '1') then							-- |A| <= n
								es0_reg	<= '0';								-- use ozp
							else
								es0_reg	<= '1';								-- not use ozp
							end if;
							ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(4,4);		-- Ek <= fix0(ozp(Pi)) or fix0(Pi)

							state_reg	<= ST_H1;
						end if;
					end if;

				when ST_H1 =>	-- (cannot merge with ST_H0; data path configuration is not yet done in H0)
					msb_a_reg		<= bdi(63);
					msb_ozp_a_reg	<= nd0(63);
					state_reg		<= ST_H2;

				-- Normal loop (|A| != 0)
				when ST_H2 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H3;
					end if;

				when ST_H3 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						bdi_ready_reg	<= '1';			-- terminate bus cycle (A)
						state_reg		<= ST_H4;
					end if;

				when ST_H4 =>
					bdi_ready_reg	<= '0';

					if (eot_reg = '1') then				-- last data
						-- reconfig data path, for calc V
						if (a1_flag_reg = '1' and msb_ozp_a_reg = '1') then
							vs0_reg	<= '0';				-- use h
						else
							vs0_reg	<= '1';				-- not use h
						end if;

						if (alen_reg = CONV_STD_LOGIC_VECTOR(8,4)) then
							vi_sel_reg	<= '0';			-- use f1
						else
							vi_sel_reg	<= '1';			-- use f2
						end if;

						state_reg	<= ST_H9;			-- calc V

					else
						state_reg	<= ST_H5;
					end if;

				when ST_H5 =>
				-- wait next bus cycle (A)
					if (bdi_valid = '1') then
						alen_reg	<= bdi_size;
						eot_reg		<= bdi_eot;
						eoi_reg		<= bdi_eoi;

						-- reconfig data path, for loading next A
						if (bdi_eot = '1') then						-- the last data
							es0_reg	<= '0';							-- use ozp
						else
							es0_reg	<= '1';							-- not use ozp
						end if;

						if (a1_flag_reg = '1' and msb_a_reg = '1') then
							es1_reg	<= '0';							-- use h
						else
							es1_reg	<= '1';							-- not use h (thru)
						end if;

						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(6,4);	-- Ek <= Pi + i/h(Eo)

						-- clear A[1] flag
						a1_flag_reg	<= '0';
						state_reg	<= ST_H2;
					end if;

				-- |A| == 0 (special case)		NOTE; possibility of bus cycle (|A| == 0 && |M| != 0)
				when ST_H6 =>	-- start Ek	(still in a bus cycle)
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H7;
					end if;

				when ST_H7 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						state_reg	<= ST_H8;
					end if;

				when ST_H8 =>	-- reconfig data path, for calc V
					vs0_reg		<= '0';							-- force to use h_func
					vi_sel_reg	<= '1';							-- force to use f2
					state_reg	<= ST_H9;						-- calc V

				-- compute and latch v_reg		NOTE; possibility of bus cycle (|A| == 0 && |M| != 0)
				when ST_H9 =>
					-- reconfig data path, for Y computation
					ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(0,4);	-- Ek <= V
					v_reg		<= vi_node;
					state_reg	<= ST_H10;

				when ST_H10 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H11;
					end if;

				when ST_H11 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						y_reg		<= ek_dout;					-- Y <= Ek(V)
						state_reg	<= ST_C0;
					end if;

				--------------------------------
				-- ENC/DEC
				--------------------------------
				-- calc X0		NOTE; possiblity of bus cycle (|A| == 0 && |M| != 0)
				when ST_C0 =>
					if (eoi_reg = '1') then							-- |M| == 0
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(1,4);	-- Ek <= g1(V)
					else
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(2,4);	-- Ek <= g2(V)
					end if;
					state_reg	<= ST_C1;

				when ST_C1 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_C2;
					end if;

				when ST_C2 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						x_reg		<= ek_dout;					-- X <= Ek(g1/g2(V))

						if (eoi_reg = '1') then
							state_reg	<= ST_P0;				-- output T
						else
							state_reg	<= ST_C3;
						end if;
					end if;

				-- |M| != 0 (normal case)	----------------------------------------/
				when ST_C3 =>	-- wait M[*] bus cycle
					if (bdi_valid = '1') then
						mlen_reg	<= bdi_size;
						eot_reg		<= bdi_eot;
						eoi_reg		<= bdi_eoi;

						-- reconfig data path, for output O
						if (bdi_eot = '1') then
							os0_reg	<= '0';						-- O <= Pi + msb(Ek)
						else
							os0_reg	<= '1';						-- O <= Pi + Ek
						end if;

						state_reg	<= ST_C4;
					end if;

				when ST_C4 =>	-- latch output
					if (enc_dec_reg = '0') then
						o_reg	<= oi_node;						-- enc
					else
						o_reg	<= bdi;							-- dec
					end if;
					od_reg			<= oi_node;					-- back up correct data (for dec)

					bdo_size_reg	<= mlen_reg;

					if (eot_reg = '0') then
						-- reconfig data path, for calc next Y
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(3,4);	-- use fix1
						state_reg	<= ST_C5;
					else
						state_reg	<= ST_C7;						-- skip computing Y
					end if;

				when ST_C5 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_C6;
					end if;

				when ST_C6 =>	-- wait end of Ek, load Y
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						y_reg		<= ek_dout;					-- Y <= Ek(fix1(O))
						state_reg	<= ST_C7;
					end if;

				when ST_C7 =>	-- reconfig data, for calc next X
					if (eot_reg = '1') then
						es2_reg		<= '0';								-- use ozp
						if (mlen_reg = CONV_STD_LOGIC_VECTOR(8,33)) then
							ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(8,4);	-- Ek <= f1(X + ozp(O))
						else
							ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(9,4);	-- Ek <= f2(X + ozp(O))
						end if;
					else
						es2_reg		<= '1';								-- not use ozp
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(7,4);		-- Ek <= X + O
					end if;

					state_reg	<= ST_C8;

				when ST_C8 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_C9;
					end if;

				when ST_C9 =>	-- wait end of Ek, load X
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						x_reg		<= ek_dout;
						state_reg	<= ST_C10;
					end if;

				-- output C/M
				when ST_C10 =>
					if (bdo_ready = '1') then
						o_reg			<= od_reg;			-- recover correct data (for dec); same data (for enc)
						bdo_valid_reg	<= '1';
						state_reg		<= ST_C11;
					end if;

				when ST_C11 =>
					bdo_valid_reg	<= '0';
					bdi_ready_reg	<= '1';					-- terminate bus cycle (M)

					if (eot_reg = '0') then
						state_reg	<= ST_C12;				-- next M
					else
						state_reg	<= ST_P0;
					end if;

				when ST_C12 =>	-- terminate bus cycle
					bdi_ready_reg	<= '0';
					state_reg		<= ST_C3;

				--------------------------------
				-- process T
				--------------------------------
				when ST_P0 =>
					bdi_ready_reg	<= '0';

					if (enc_dec_reg = '0') then
						state_reg	<= ST_P1;				-- ENC
					else
						state_reg	<= ST_P3;				-- DEC
					end if;

				-- ENC
				-- output T/result
				when ST_P1 =>	-- issue output T/result
					if (bdo_ready = '1') then
						o_reg			<= (t_node & CONV_STD_LOGIC_VECTOR(0,32));
						bdo_valid_reg	<= '1';
						state_reg		<= ST_P2;
					end if;

				when ST_P2 =>	-- wait end of bus cycle
					bdo_valid_reg	<= '0';
					state_reg		<= ST_WAIT;

				-- DEC
				when ST_P3 =>	-- T
					if (bdi_valid = '1') then
						-- check T
						if (bdi(63 downto 32) = t_node) then	-- TAU == 4
							msg_auth_valid_reg	<= '1';
						else
							msg_auth_valid_reg	<= '0';
						end if;
						msg_auth_done_reg	<= '1';

						state_reg			<= ST_P4;
					end if;

				when ST_P4 =>
					msg_auth_done_reg	<= '0';
					msg_auth_valid_reg	<= '0';

					bdi_ready_reg	<= '1';			-- terminate bus cycle (T)
					state_reg		<= ST_P5;

				when ST_P5 =>
					bdi_ready_reg	<= '0';
					state_reg		<= ST_WAIT;

				when others =>
					null;
				end case;
			end if;
		end if;
	end process;

	------------------------------------------------------------/
	-- PO
	------------------------------------------------------------/
	bdi_ready	<= bdi_ready_reg;
	key_ready	<= key_ready_reg;
	bdo_valid	<= bdo_valid_reg;
	bdo			<= o_reg;

	bdo_size		<= bdo_size_reg;
	msg_auth_done	<= msg_auth_done_reg;
	msg_auth_valid	<= msg_auth_valid_reg;

end RTL;
