------------------------------------------------------------------------------
-- Twine (80bit key, enc only)
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity twine80 is
	port (
		din		: in  std_logic_vector(63 downto 0);
		dout	: out std_logic_vector(63 downto 0);
		kin		: in  std_logic_vector(79 downto 0);

		start	: in  std_logic;
		busy	: out std_logic;

		rst		: in  std_logic;	-- sync reset, active H
		clk		: in  std_logic
	);
end twine80;

architecture RTL of twine80 is
	constant ROUND_PER_CLK	: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(1,6);

	------------------------------------------------------------/
	-- internal wires/registers
	------------------------------------------------------------/
	signal data_reg			: std_logic_vector(63 downto 0);
	signal round_key_reg	: std_logic_vector(79 downto 0);

	signal next_data		: std_logic_vector(63 downto 0);
	signal next_key			: std_logic_vector(79 downto 0);
	signal next_rc			: std_logic_vector(5 downto 0);
	signal last_round		: std_logic;
	signal next_round_val	: std_logic_vector(214 downto 0);

	constant RC62			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"21")),6);	-- (== C^-1)
	constant RC0			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"01")),6);
	constant RC1			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"02")),6);
	constant RC2			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"04")),6);
	constant RC30			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"33")),6);
	constant RC31			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"25")),6);
	constant RC32			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"09")),6);
	constant RC33			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"12")),6);
	constant RC34			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"24")),6);
	constant RC35			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"0b")),6);
	constant RC36			: std_logic_vector(5 downto 0) := CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"16")),6);
	signal rc_reg			: std_logic_vector(5 downto 0);

	constant ST_WAIT		: std_logic := '0';
	constant ST_CALC0		: std_logic := '1';
	signal state_reg		: std_logic;

	------------------------------------------------------------/
	-- functions
	------------------------------------------------------------/

	----------------
	-- S
	----------------
	function s_func (
		i	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable o			: std_logic_vector(3 downto 0);
		variable s_func_ret	: std_logic_vector(3 downto 0);
	begin
		case (i) is
		when To_StdLogicVector(X"0") =>
			o	:= To_StdLogicVector(X"C");
		when To_StdLogicVector(X"1") =>
			o	:= To_StdLogicVector(X"0");
		when To_StdLogicVector(X"2") =>
			o	:= To_StdLogicVector(X"F");
		when To_StdLogicVector(X"3") =>
			o	:= To_StdLogicVector(X"A");
		when To_StdLogicVector(X"4") =>
			o	:= To_StdLogicVector(X"2");
		when To_StdLogicVector(X"5") =>
			o	:= To_StdLogicVector(X"B");
		when To_StdLogicVector(X"6") =>
			o	:= To_StdLogicVector(X"9");
		when To_StdLogicVector(X"7") =>
			o	:= To_StdLogicVector(X"5");
		when To_StdLogicVector(X"8") =>
			o	:= To_StdLogicVector(X"8");
		when To_StdLogicVector(X"9") =>
			o	:= To_StdLogicVector(X"3");
		when To_StdLogicVector(X"A") =>
			o	:= To_StdLogicVector(X"D");
		when To_StdLogicVector(X"B") =>
			o	:= To_StdLogicVector(X"7");
		when To_StdLogicVector(X"C") =>
			o	:= To_StdLogicVector(X"1");
		when To_StdLogicVector(X"D") =>
			o	:= To_StdLogicVector(X"E");
		when To_StdLogicVector(X"E") =>
			o	:= To_StdLogicVector(X"6");
		when others =>
			o	:= To_StdLogicVector(X"4");
		end case;

		s_func_ret	:= o;
		return s_func_ret;
	end s_func;

	----------------
	-- Fi
	----------------
	function f_func (
		x	: std_logic_vector(3 downto 0);
		ek	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable f_func_ret	: std_logic_vector(3 downto 0);
	begin
		f_func_ret	:= s_func(x xor ek);
		return f_func_ret;
	end f_func;

	----------------
	-- P
	----------------
	function p_func (
		i	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable x0, x1, x2, x3, x4, x5, x6, x7			: std_logic_vector(3 downto 0);
		variable x8, x9, x10, x11, x12, x13, x14, x15	: std_logic_vector(3 downto 0);
		variable p_func_ret								: std_logic_vector(63 downto 0);
	begin
		x0	:= i(63 downto 60);
		x1	:= i(59 downto 56);
		x2	:= i(55 downto 52);
		x3	:= i(51 downto 48);
		x4	:= i(47 downto 44);
		x5	:= i(43 downto 40);
		x6	:= i(39 downto 36);
		x7	:= i(35 downto 32);
		x8	:= i(31 downto 28);
		x9	:= i(27 downto 24);
		x10	:= i(23 downto 20);
		x11	:= i(19 downto 16);
		x12	:= i(15 downto 12);
		x13	:= i(11 downto 8);
		x14	:= i(7 downto 4);
		x15	:= i(3 downto 0);

		p_func_ret	:= (x1 & x2 & x11 & x6 & x3 & x0 & x9 & x4 & x7 & x10 & x13 & x14 & x5 & x8 & x15 & x12);
		return p_func_ret;
	end p_func;

	----------------
	-- Round
	----------------
	function round_func (
		p			: std_logic_vector(63 downto 0);
		ek			: std_logic_vector(31 downto 0);
		last_round	: std_logic
	) return std_logic_vector is
		variable p0, p1, p2, p3, p4, p5, p6, p7			: std_logic_vector(3 downto 0);
		variable p8, p9, p10, p11, p12, p13, p14, p15	: std_logic_vector(3 downto 0);
		variable ek0, ek1, ek2, ek3, ek4, ek5, ek6, ek7	: std_logic_vector(3 downto 0);
		variable f										: std_logic_vector(63 downto 0);
		variable pi										: std_logic_vector(63 downto 0);
		variable round_func_ret							: std_logic_vector(127 downto 0);
	begin
		p0	:= p(63 downto 60);
		p1	:= p(59 downto 56);
		p2	:= p(55 downto 52);
		p3	:= p(51 downto 48);
		p4	:= p(47 downto 44);
		p5	:= p(43 downto 40);
		p6	:= p(39 downto 36);
		p7	:= p(35 downto 32);
		p8	:= p(31 downto 28);
		p9	:= p(27 downto 24);
		p10	:= p(23 downto 20);
		p11	:= p(19 downto 16);
		p12	:= p(15 downto 12);
		p13	:= p(11 downto 8);
		p14	:= p(7 downto 4);
		p15	:= p(3 downto 0);

		ek0	:= ek(31 downto 28);
		ek1	:= ek(27 downto 24);
		ek2	:= ek(23 downto 20);
		ek3	:= ek(19 downto 16);
		ek4	:= ek(15 downto 12);
		ek5	:= ek(11 downto 8);
		ek6	:= ek(7 downto 4);
		ek7	:= ek(3 downto 0);

		f	:= (p0 & (p1 xor f_func(p0, ek0)) & p2 & (p3 xor f_func(p2, ek1)) 
					& p4 & (p5 xor f_func(p4, ek2)) & p6 & (p7 xor f_func(p6, ek3)) 
					& p8 & (p9 xor f_func(p8, ek4)) & p10 & (p11 xor f_func(p10, ek5)) 
					& p12 & (p13 xor f_func(p12, ek6)) & p14 & (p15 xor f_func(p14, ek7)) );

		round_func_ret	:= (f & p_func(f));
		return round_func_ret;
	end round_func;

	----------------
	-- key scheduler
	----------------
	function key_func (
		k	: std_logic_vector(79 downto 0);
		c	: std_logic_vector(5 downto 0)
	) return std_logic_vector is
		variable k0, k1, k2, k3		: std_logic_vector(3 downto 0);
		variable k4, k5, k6, k7		: std_logic_vector(3 downto 0);
		variable k8, k9, k10, k11	: std_logic_vector(3 downto 0);
		variable k12, k13, k14, k15	: std_logic_vector(3 downto 0);
		variable k16, k17, k18, k19	: std_logic_vector(3 downto 0);
		variable nk					: std_logic_vector(79 downto 0);
		variable key_func_ret		: std_logic_vector(79 downto 0);
	begin
		k0	:= k(79 downto 76);
		k1	:= k(75 downto 72);
		k2	:= k(71 downto 68);
		k3	:= k(67 downto 64);
		k4	:= k(63 downto 60);
		k5	:= k(59 downto 56);
		k6	:= k(55 downto 52);
		k7	:= k(51 downto 48);
		k8	:= k(47 downto 44);
		k9	:= k(43 downto 40);
		k10	:= k(39 downto 36);
		k11	:= k(35 downto 32);
		k12	:= k(31 downto 28);
		k13	:= k(27 downto 24);
		k14	:= k(23 downto 20);
		k15	:= k(19 downto 16);
		k16	:= k(15 downto 12);
		k17	:= k(11 downto 8);
		k18	:= k(7 downto 4);
		k19	:= k(3 downto 0);

		nk(79 downto 76)	:= k1 xor s_func(k0);
		nk(75 downto 72)	:= k2;
		nk(71 downto 68)	:= k3;
		nk(67 downto 64)	:= k0;
		nk(63 downto 60)	:= k4 xor s_func(k16);
		nk(59 downto 56)	:= k5;
		nk(55 downto 52)	:= k6;
		nk(51 downto 48)	:= k7 xor ('0' & c(5 downto 3));
		nk(47 downto 44)	:= k8;
		nk(43 downto 40)	:= k9;
		nk(39 downto 36)	:= k10;
		nk(35 downto 32)	:= k11;
		nk(31 downto 28)	:= k12;
		nk(27 downto 24)	:= k13;
		nk(23 downto 20)	:= k14;
		nk(19 downto 16)	:= k15;
		nk(15 downto 12)	:= k16;
		nk(11 downto 8)	:= k17;
		nk(7 downto 4)	:= k18;
		nk(3 downto 0)	:= k19 xor ('0' & c(2 downto 0));

		key_func_ret(79 downto 16)	:= nk(63 downto 0);
		key_func_ret(15 downto 0)	:= nk(79 downto 64);
		return key_func_ret;
	end key_func;

	----------------
	-- RC
	----------------
	function rc_func (
		k	: std_logic_vector(5 downto 0)
	) return std_logic_vector is
		variable rc_func_ret	: std_logic_vector(5 downto 0);
	begin
		-- a^1 = 000010
		rc_func_ret	:= (k(4) & k(3) & k(2) & k(1) & (k(0) xor k(5)) & k(5));
		return rc_func_ret;
	end rc_func;

	----------------
	-- multiple rounds
	----------------
	function multi_round_func (
		data	: std_logic_vector(63 downto 0);
		key		: std_logic_vector(79 downto 0);
		rc		: std_logic_vector(5 downto 0)
	) return std_logic_vector is
		variable cd						: std_logic_vector(63 downto 0);
		variable nd						: std_logic_vector(127 downto 0);
		variable ck, nk					: std_logic_vector(79 downto 0);
		variable cr, nr					: std_logic_vector(5 downto 0);
		variable ek						: std_logic_vector(31 downto 0);
		variable last					: std_logic;
		variable multi_round_func_ret	: std_logic_vector(214 downto 0);
	begin
		cd	:= data;
		ck	:= key;
		cr	:= rc;

		for I in 0 to CONV_INTEGER(ROUND_PER_CLK) - 1 loop
			if (cr = RC34) then
				last	:= '1';
			else
				last	:= '0';
			end if;

			ek	:= (ck(75 downto 72) & ck(67 downto 60) & ck(55 downto 52) & ck(27 downto 12));
			nd	:= round_func(cd, ek, last);
			nk	:= key_func(ck, cr);
			nr	:= rc_func(cr);

			cd	:= nd(63 downto 0);
			ck	:= nk;
			cr	:= nr;
		end loop;

		multi_round_func_ret	:= (nd & ck & cr & last);
		return multi_round_func_ret;
	end multi_round_func;

begin
	next_round_val	<= multi_round_func(data_reg, round_key_reg, rc_reg);
	next_data		<= next_round_val(150 downto 87);
	next_key		<= next_round_val(86 downto 7);
	next_rc			<= next_round_val(6 downto 1);
	last_round		<= next_round_val(0);

	------------------------------------------------------------/
	-- Main control
	------------------------------------------------------------/
	p0:process (clk)
	begin
		if (clk'event and clk = '1') then
			if (rst = '1') then
				state_reg	<= ST_WAIT;

			else
				case (state_reg) is
				--------------------------------
				-- wait trigger
				--------------------------------
				when ST_WAIT =>
					if (start = '1') then
						round_key_reg	<= kin;
						data_reg		<= din;
						rc_reg			<= RC0;
						state_reg		<= ST_CALC0;
					end if;

				--------------------------------
				-- ENC
				--------------------------------
				when ST_CALC0 =>
					data_reg		<= next_data;
					round_key_reg	<= next_key;
					rc_reg			<= next_rc;

					if (last_round = '1') then
						state_reg	<= ST_WAIT;
					end if;

				when others =>
					null;
				end case;
			end if;
		end if;
	end process;

	------------------------------------------------------------/
	-- PO
	------------------------------------------------------------/
	dout	<= next_round_val(214 downto 151);
	busy	<= state_reg;

end RTL;
