------------------------------------------------------------------------------
-- SILC-LED80 with AEAD I/F
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity silc_led80_aeadif is
	port (
		bdi				: in  std_logic_vector(63 downto 0);
		bdi_valid		: in  std_logic;
		bdi_ready		: out std_logic;
		decrypt			: in  std_logic;
		bdi_type		: in  std_logic_vector(2 downto 0);
		bdi_partial		: in  std_logic;
		bdi_eot			: in  std_logic;
		bdi_eoi			: in  std_logic;
		bdi_size		: in  std_logic_vector(3 downto 0);
		bdi_valid_bytes	: in  std_logic_vector(7 downto 0);
		bdi_pad_loc		: in  std_logic_vector(7 downto 0);

		key				: in  std_logic_vector(79 downto 0);
		key_valid		: in  std_logic;
		key_update		: in  std_logic;
		key_ready		: out std_logic;

		bdo				: out std_logic_vector(63 downto 0);
		bdo_valid		: out std_logic;
		bdo_ready		: in  std_logic;
		bdo_size		: out std_logic_vector(3 downto 0);
		msg_auth_done	: out std_logic;
		msg_auth_valid	: out std_logic;

		rst				: in  std_logic;	-- sync reset, active H
		clk				: in  std_logic
	);
end silc_led80_aeadif;

architecture RTL of silc_led80_aeadif is

	component led80
	port (
		din		: in  std_logic_vector(63 downto 0);
		dout	: out std_logic_vector(63 downto 0);
		kin		: in  std_logic_vector(79 downto 0);
		start	: in  std_logic;
		busy	: out std_logic;
		rst		: in  std_logic;
		clk		: in  std_logic
	);
	end component;

	constant PARAM	: std_logic_vector(7 downto 0)  := To_StdLogicVector(X"C8");		-- led t6n4
	constant TAU	: std_logic_vector(31 downto 0) := CONV_STD_LOGIC_VECTOR(4,32);

	------------------------------------------------------------/
	-- functions
	------------------------------------------------------------/

	-- zpp
	function zpp_func (
		a	: std_logic_vector(63 downto 0);
		len	: std_logic_vector(3 downto 0)	-- valid length (unit:byte, 0-8)
	) return std_logic_vector is
		variable zpp_func_ret	: std_logic_vector(63 downto 0);
	begin
		case (len) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			zpp_func_ret	:= CONV_STD_LOGIC_VECTOR(0,64);
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,56) & a(63 downto 56));
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,48) & a(63 downto 48));
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,40) & a(63 downto 40));
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,32) & a(63 downto 32));
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,24) & a(63 downto 24));
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,16) & a(63 downto 16));
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			zpp_func_ret	:= (CONV_STD_LOGIC_VECTOR(0,8) & a(63 downto 8));
		when others =>	-- >=8; thru
			zpp_func_ret	:= a;
		end case;
		return zpp_func_ret;
	end zpp_func;

	-- zap
	function zap_func (
		a		: std_logic_vector(63 downto 0);
		len		: std_logic_vector(3 downto 0);	-- valid length (unit:byte, 0-8)
		thru	: std_logic
	) return std_logic_vector is
		variable and_mask		: std_logic_vector(63 downto 0);
		variable len_r			: std_logic_vector(3 downto 0);
		variable zap_func_ret	: std_logic_vector(63 downto 0);
	begin
		if (thru = '1') then
			len_r	:= CONV_STD_LOGIC_VECTOR(8,4);
		else
			len_r	:= len;
		end if;

		case (len_r) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			and_mask	:= To_StdLogicVector(X"0000000000000000");
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			and_mask	:= To_StdLogicVector(X"FF00000000000000");
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			and_mask	:= To_StdLogicVector(X"FFFF000000000000");
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFF0000000000");
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFF00000000");
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFF000000");
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFF0000");
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFF00");
		when others =>	-- >=8; thru
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFFFF");
		end case;

		zap_func_ret	:= a and and_mask;
		return zap_func_ret;
	end zap_func;

	-- fix1
	function fix1_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable fix1_func_ret	: std_logic_vector(63 downto 0);
	begin
		fix1_func_ret	:= ('1' & a(62 downto 0));
		return fix1_func_ret;
	end fix1_func;

	-- g
	function g_func (
		a	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable g_func_ret	: std_logic_vector(63 downto 0);
	begin
		g_func_ret	:= (a(55 downto 0) & (a(63 downto 56) xor a(55 downto 48)) );
		return g_func_ret;
	end g_func;

	-- msb
	function msb_func (
		a		: std_logic_vector(63 downto 0);
		len		: std_logic_vector(3 downto 0);	-- valid length (unit:byte, 0-8)
		thru	: std_logic
	) return std_logic_vector is
		variable and_mask		: std_logic_vector(63 downto 0);
		variable len_r			: std_logic_vector(3 downto 0);
		variable msb_func_ret	: std_logic_vector(63 downto 0);
	begin
		if (thru = '1') then
			len_r	:= CONV_STD_LOGIC_VECTOR(8,4);
		else
			len_r	:= len;
		end if;

		case (len_r) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			and_mask	:= To_StdLogicVector(X"0000000000000000");
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			and_mask	:= To_StdLogicVector(X"FF00000000000000");
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			and_mask	:= To_StdLogicVector(X"FFFF000000000000");
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFF0000000000");
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFF00000000");
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFF000000");
		when CONV_STD_LOGIC_VECTOR(6,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFF0000");
		when CONV_STD_LOGIC_VECTOR(7,4) =>
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFF00");
		when others =>	-- >=8; thru
			and_mask	:= To_StdLogicVector(X"FFFFFFFFFFFFFFFF");
		end case;

		msb_func_ret	:= a and and_mask;
		return msb_func_ret;
	end msb_func;

	-- mux
	function mux7_func (
		a0	: std_logic_vector(63 downto 0);
		a1	: std_logic_vector(63 downto 0);
		a2	: std_logic_vector(63 downto 0);
		a3	: std_logic_vector(63 downto 0);
		a4	: std_logic_vector(63 downto 0);
		a5	: std_logic_vector(63 downto 0);
		a6	: std_logic_vector(63 downto 0);
		sel	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable mux7_func_ret	: std_logic_vector(63 downto 0);
	begin
		case (sel) is
		when CONV_STD_LOGIC_VECTOR(0,4) =>
			mux7_func_ret	:= a0;
		when CONV_STD_LOGIC_VECTOR(1,4) =>
			mux7_func_ret	:= a1;
		when CONV_STD_LOGIC_VECTOR(2,4) =>
			mux7_func_ret	:= a2;
		when CONV_STD_LOGIC_VECTOR(3,4) =>
			mux7_func_ret	:= a3;
		when CONV_STD_LOGIC_VECTOR(4,4) =>
			mux7_func_ret	:= a4;
		when CONV_STD_LOGIC_VECTOR(5,4) =>
			mux7_func_ret	:= a5;
		when others =>
			mux7_func_ret	:= a6;
		end case;
		return mux7_func_ret;
	end mux7_func;

	------------------------------------------------------------/
	-- Main data path
	------------------------------------------------------------/
	signal nonce_reg							: std_logic_vector(55 downto 0);	-- NOTE; TAU==4
	signal key_reg								: std_logic_vector(79 downto 0);
	signal v_reg, x_reg, y_reg, o_reg, od_reg	: std_logic_vector(63 downto 0);
	signal alen_reg, mlen_reg, nlen_reg			: std_logic_vector(3 downto 0);
	signal alen_org_reg, mlen_org_reg			: std_logic_vector(63 downto 0);
	signal pi_par								: std_logic_vector(63 downto 0);

	signal ek_din, ek_dout						: std_logic_vector(63 downto 0);
	signal ek_start_reg							: std_logic;
	signal ek_busy								: std_logic;

	-- Ei
	signal ei_sel_reg							: std_logic_vector(3 downto 0);	-- data path control
	signal es0_reg, es1_reg						: std_logic;					-- data path control
	signal ei0, ei1, ei2, ei3, ei4, ei5, ei6	: std_logic_vector(63 downto 0);

	-- Vi
	signal vi_node								: std_logic_vector(63 downto 0);

	-- Oi
	signal os0_reg								: std_logic;
	signal oi_node								: std_logic_vector(63 downto 0);

	-- T
	signal t_check_reg							: std_logic;
	signal t_node								: std_logic_vector(31 downto 0);	-- TAU == 4

	------------------------------------------------------------/
	-- Main control
	------------------------------------------------------------/
	signal bdi_ready_reg, key_ready_reg, bdo_valid_reg	: std_logic;
	signal enc_dec_reg									: std_logic;

	signal bdo_size_reg									: std_logic_vector(3 downto 0);
	signal msg_auth_done_reg							: std_logic;
	signal msg_auth_valid_reg							: std_logic;
	signal eot_reg, eoi_reg								: std_logic;
	signal no_a_reg										: std_logic;

	constant ST_WAIT	: std_logic_vector(7 downto 0) := To_StdLogicVector(X"00");
	constant ST_K0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"10");
	constant ST_N0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"20");
	constant ST_H0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"30");
	constant ST_H1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"31");
	constant ST_H2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"32");
	constant ST_H3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"33");
	constant ST_H4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"34");
	constant ST_H5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"35");
	constant ST_H6		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"36");
	constant ST_H7		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"37");
	constant ST_H8		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"38");
	constant ST_H9		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"39");
	constant ST_H10		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"40");
	constant ST_H11		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"41");
	constant ST_H12		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"42");
	constant ST_C0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"50");
	constant ST_C1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"51");
	constant ST_C2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"52");
	constant ST_C3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"53");
	constant ST_C4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"54");
	constant ST_C5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"55");
	constant ST_C6		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"56");
	constant ST_C7		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"57");
	constant ST_C8		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"58");
	constant ST_C9		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"59");
	constant ST_P0		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"70");
	constant ST_P1		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"71");
	constant ST_P2		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"72");
	constant ST_P3		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"73");
	constant ST_P4		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"74");
	constant ST_P5		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"75");
	constant ST_P6		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"76");
	constant ST_P7		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"77");
	constant ST_P8		: std_logic_vector(7 downto 0) := To_StdLogicVector(X"78");

	signal state_reg	: std_logic_vector(7 downto 0);

begin
	------------------------------------------------------------/
	-- Main data path
	------------------------------------------------------------/
	pi_par	<= (PARAM & nonce_reg);	-- NOTE; TAU==4

	ei0		<= v_reg;
	ei1		<= g_func(v_reg);
	ei2		<= g_func(mlen_org_reg xor x_reg);
	ei3		<= fix1_func(o_reg);
	ei4		<= zpp_func(pi_par, nlen_reg(3 downto 0));
	ei5		<= zap_func(bdi, alen_reg(3 downto 0), es0_reg) xor ek_dout;
	ei6		<= zap_func(o_reg, mlen_reg(3 downto 0), es1_reg) xor x_reg;

	ek_din	<= mux7_func(ei0, ei1, ei2, ei3, ei4, ei5, ei6, ei_sel_reg);

	vi_node	<= g_func(alen_org_reg xor ek_dout);

	oi_node	<= msb_func(y_reg, mlen_reg(3 downto 0), os0_reg) xor bdi;

	t_node	<= x_reg(63 downto 32);

	-- Ek
	u0: led80
	port map (
		din		=> ek_din,
		dout	=> ek_dout,
		kin		=> key_reg,
		start	=> ek_start_reg,
		busy	=> ek_busy,
		rst		=> rst,
		clk		=> clk
	);

	------------------------------------------------------------/
	-- Main control
	------------------------------------------------------------/
	pr0:process (clk)
	begin
		if (clk'event and clk = '1') then
			if (rst = '1') then
				bdi_ready_reg		<= '0';
				key_ready_reg		<= '0';
				bdo_valid_reg		<= '0';
				ek_start_reg		<= '0';

				bdo_size_reg		<= CONV_STD_LOGIC_VECTOR(0,4);
				msg_auth_done_reg	<= '0';
				msg_auth_valid_reg	<= '0';
				eot_reg				<= '0';
				eoi_reg				<= '0';
				no_a_reg			<= '0';

				ei_sel_reg			<= CONV_STD_LOGIC_VECTOR(0,4);
				es0_reg				<= '0';
				es1_reg				<= '0';
				os0_reg				<= '0';

				state_reg			<= ST_WAIT;

			else
				case (state_reg) is
				--------------------------------
				-- wait trigger
				--------------------------------
				when ST_WAIT =>
					if (key_update = '1') then
						if (key_valid = '1') then
							key_reg			<= key;
							key_ready_reg	<= '1';
							state_reg		<= ST_K0;
						end if;

					elsif (bdi_valid = '1' and bdi_type = CONV_STD_LOGIC_VECTOR(6,3)) then
						nlen_reg		<= bdi_size + '1';			-- [add 1, which corresponds to length of const value]
						nonce_reg		<= bdi(63 downto 8);		-- ignore last 8bit (for adding a constant)
						enc_dec_reg		<= decrypt;
						eot_reg			<= bdi_eot;
						eoi_reg			<= bdi_eoi;

						bdi_ready_reg	<= '1';						-- terminate bus cycle (nonce)
						state_reg		<= ST_N0;

					else
						key_ready_reg	<= '0';
						bdi_ready_reg	<= '0';
					end if;

				--------------------------------
				-- SDI: Set key
				--------------------------------
				when ST_K0 =>	-- wait end of bus cycle (key)
					key_ready_reg	<= '0';
					state_reg		<= ST_WAIT;

				--------------------------------
				-- PDI1: NONCE
				--------------------------------
				when ST_N0 =>
					bdi_ready_reg	<= '0';

					-- reconfig data path
					ei_sel_reg		<= CONV_STD_LOGIC_VECTOR(4,4);	-- Ek <= zpp(param)

					alen_reg		<= CONV_STD_LOGIC_VECTOR(0,4);
					alen_org_reg	<= CONV_STD_LOGIC_VECTOR(0,64);

					if (eoi_reg = '0') then
						state_reg	<= ST_H0;
					else
						no_a_reg	<= '1';							-- |A| = 0, |M| = 0; skip bus wait
						state_reg	<= ST_H1;						-- special case
					end if;

				--------------------------------
				-- PD2: HASH
				--------------------------------
				when ST_H0 =>	-- wait bus cycle; A[1] or M[1]
					if (bdi_valid = '1') then
						if (bdi_type(2 downto 1) = "01") then		-- M/C comes here, indicating |A| = 0
							-- NOTE: do not terminate this bus cycle (M/C)
							alen_reg		<= CONV_STD_LOGIC_VECTOR(0,4);
							alen_org_reg	<= CONV_STD_LOGIC_VECTOR(0,64);

							no_a_reg		<= '1';

						else
							alen_reg		<= bdi_size;			-- (normal case)
							alen_org_reg	<= alen_org_reg + bdi_size;
							eot_reg			<= bdi_eot;
							eoi_reg			<= bdi_eoi;

							no_a_reg		<= '0';
						end if;

						state_reg	<= ST_H1;
					end if;

				when ST_H1 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H2;
					end if;

				when ST_H2 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						if (no_a_reg = '1') then
							state_reg	<= ST_H8;					-- (special case)
						else
							state_reg	<= ST_H3;
						end if;
					end if;

				-- Normal loop (|A| != 0)
				when ST_H3 =>	-- reconfig datapath
					if (eot_reg = '1') then							-- last data
						es0_reg	<= '0';								-- Ek <= zap(Pi) + Ek
					else
						es0_reg	<= '1';								-- Ek <= Pi + Ek
					end if;
					ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(5,4);

					state_reg	<= ST_H4;

				when ST_H4 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H5;
					end if;

				when ST_H5 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						bdi_ready_reg	<= '1';						-- terminate bus cycle (A)
						state_reg		<= ST_H6;
					end if;

				when ST_H6 =>
					bdi_ready_reg	<= '0';

					if (eot_reg = '1') then							-- last data
						state_reg	<= ST_H8;
					else
						state_reg	<= ST_H7;
					end if;

				when ST_H7 =>	-- wait next bus cycle (A)
					if (bdi_valid = '1') then
						alen_reg		<= bdi_size;
						alen_org_reg	<= alen_org_reg + bdi_size;
						eot_reg			<= bdi_eot;
						eoi_reg			<= bdi_eoi;

						state_reg		<= ST_H3;
					end if;

				-- latch v_reg	NOTE; possibility of bus cycle (|A| == 0 && |M| != 0)
				when ST_H8 =>
					v_reg		<= vi_node;
					ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(1,4);
					state_reg	<= ST_H9;

				-- calc X
				when ST_H9 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H10;
					end if;

				when ST_H10 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						x_reg		<= ek_dout;						-- X <= Ek(g(V))
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(0,4);
						state_reg	<= ST_H11;
					end if;

				-- calc Y
				when ST_H11 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_H12;
					end if;

				when ST_H12 =>	-- wait end of Ek
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						y_reg			<= ek_dout;					-- Y <= Ek(V)

						mlen_reg		<= CONV_STD_LOGIC_VECTOR(0,4);
						mlen_org_reg	<= CONV_STD_LOGIC_VECTOR(0,64);

						if (eoi_reg = '1') then						-- No M/C data
							state_reg	<= ST_P0;
						else
							state_reg	<= ST_C0;
						end if;
					end if;

				--------------------------------
				-- ENC/DEC
				--------------------------------

				-- |M| != 0 (normal case)	----------------------------------------
				when ST_C0 =>	-- HEADER
					if (bdi_valid = '1') then
						mlen_reg		<= bdi_size;
						mlen_org_reg	<= mlen_org_reg + bdi_size;
						eot_reg			<= bdi_eot;
						eoi_reg			<= bdi_eoi;

						-- reconfig data path, for output O
						if (bdi_eot = '1') then
							os0_reg	<= '0';							-- O <= Pi + msb(Y)
						else
							os0_reg	<= '1';							-- O <= Pi + Y
						end if;

						state_reg	<= ST_C1;
					end if;

				when ST_C1 =>	-- latch output
					if (enc_dec_reg = '0') then
						o_reg	<= oi_node;							-- enc
					else
						o_reg	<= bdi;								-- dec
					end if;
					od_reg			<= oi_node;						-- back up correct data (for dec)

					bdo_size_reg	<= mlen_reg;

					if (eot_reg = '0') then
						-- reconfig data path, for calc next Y
						ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(3,4);	-- use fix1
						state_reg	<= ST_C2;
					else
						state_reg	<= ST_C4;						-- skip computing Y
					end if;

				when ST_C2 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_C3;
					end if;

				when ST_C3 =>	-- wait end of Ek, load Y
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						y_reg		<= ek_dout;						-- Y <= Ek(fix1(O))
						state_reg	<= ST_C4;
					end if;

				when ST_C4 =>
					-- reconfig data, for calc next X
					if (eot_reg = '1') then
						es1_reg	<= '0';								-- use zap
					else
						es1_reg	<= '1';								-- not use zap
					end if;
					ei_sel_reg	<= CONV_STD_LOGIC_VECTOR(6,4);		-- X + O or zap(O)

					state_reg	<= ST_C5;

				when ST_C5 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_C6;
					end if;

				when ST_C6 =>	-- wait end of Ek, load X
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						x_reg		<= ek_dout;
						state_reg	<= ST_C7;
					end if;

				-- output C/M
				when ST_C7 =>
					if (bdo_ready = '1') then
						o_reg			<= od_reg;					-- recover correct data (for dec); same data (for enc)
						bdo_valid_reg	<= '1';
						state_reg		<= ST_C8;
					end if;

				when ST_C8 =>
					bdo_valid_reg	<= '0';
					bdi_ready_reg	<= '1';							-- terminate bus cycle (M)

					if (eot_reg = '0') then
						state_reg	<= ST_C9;						-- next M
					else
						state_reg	<= ST_P0;
					end if;

				when ST_C9 =>	-- terminate bus cycle
					bdi_ready_reg	<= '0';
					state_reg		<= ST_C0;

				--------------------------------
				-- process T
				--------------------------------
				-- calc T
				when ST_P0 =>
					bdi_ready_reg	<= '0';

					-- reconfig data path
					ei_sel_reg		<= CONV_STD_LOGIC_VECTOR(2,4);	-- Ek <= g(X + Len C))
					state_reg		<= ST_P1;

				when ST_P1 =>	-- start Ek
					ek_start_reg	<= '1';
					if (ek_busy = '1') then
						state_reg	<= ST_P2;
					end if;

				when ST_P2 =>	-- wait end of Ek, load X
					ek_start_reg	<= '0';
					if (ek_busy = '0') then
						x_reg		<= ek_dout;						-- X
						state_reg	<= ST_P3;
					end if;

				when ST_P3 =>
					if (enc_dec_reg = '0') then
						state_reg	<= ST_P4;						-- ENC
					else
						state_reg	<= ST_P6;						-- DEC
					end if;

				-- ENC
				-- output T/result
				when ST_P4 =>	-- issue output T/result
					if (bdo_ready = '1') then
						o_reg			<= (t_node & CONV_STD_LOGIC_VECTOR(0,32));
						bdo_valid_reg	<= '1';
						state_reg		<= ST_P5;
					end if;

				when ST_P5 =>	-- wait end of bus cycle
					bdo_valid_reg	<= '0';
					state_reg		<= ST_WAIT;

				-- DEC
				-- read T (dec only)
				when ST_P6 =>	-- T
					if (bdi_valid = '1') then
						-- check T
						if (bdi(63 downto 32) = t_node) then		-- TAU == 4
							msg_auth_valid_reg	<= '1';
						else
							msg_auth_valid_reg	<= '0';
						end if;
						msg_auth_done_reg	<= '1';

						state_reg			<= ST_P7;
					end if;

				when ST_P7 =>
					msg_auth_done_reg	<= '0';
					msg_auth_valid_reg	<= '0';

					bdi_ready_reg		<= '1';						-- terminate bus cycle (T)
					state_reg			<= ST_P8;

				when ST_P8 =>
					bdi_ready_reg		<= '0';
					state_reg			<= ST_WAIT;

				when others =>
					null;
				end case;
			end if;
		end if;
	end process;

	------------------------------------------------------------/
	-- PO
	------------------------------------------------------------/
	bdi_ready	<= bdi_ready_reg;
	key_ready	<= key_ready_reg;
	bdo_valid	<= bdo_valid_reg;
	bdo			<= o_reg;

	bdo_size		<= bdo_size_reg;
	msg_auth_done	<= msg_auth_done_reg;
	msg_auth_valid	<= msg_auth_valid_reg;

end RTL;
